# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2022 PANTHEON.tech s.r.o.
# Copyright(c) 2022 University of New Hampshire

"""Libraries and utilities for running DPDK Test Suite (DTS).

The various modules in the DTS framework offer:

* Connections to nodes, both interactive and non-interactive,
* A straightforward way to add support for different operating systems of remote nodes,
* Test suite setup, execution and teardown, along with test case setup, execution and teardown,
* Pre-test suite setup and post-test suite teardown.
"""
