/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2024 Realtek Corporation. All rights reserved
 */

#ifndef RTL8125A_MCU_H
#define RTL8125A_MCU_H

void rtl_set_mac_mcu_8125a_1(struct rtl_hw *hw);
void rtl_set_mac_mcu_8125a_2(struct rtl_hw *hw);

void rtl_set_phy_mcu_8125a_1(struct rtl_hw *hw);
void rtl_set_phy_mcu_8125a_2(struct rtl_hw *hw);

#endif /* RTL8125A_MCU_H */
