/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#ifndef NTHW_REGISTER_H_
#define NTHW_REGISTER_H_

#include <unistd.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>

#include "nthw_fpga_model.h"

#include "fpga_model.h"

#include "nthw_fpga_mod_defs.h"
#include "nthw_fpga_param_defs.h"
#include "nthw_fpga_reg_defs.h"

#endif	/* NTHW_REGISTER_H_ */
