/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#ifndef __NTHW_CORE_H__
#define __NTHW_CORE_H__

#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "nthw_platform_drv.h"
#include "nthw_fpga_model.h"
#include "nthw_hif.h"
#include "nthw_pcie3.h"
#include "nthw_iic.h"
#include "nthw_i2cm.h"

#include "nthw_gmf.h"
#include "nthw_gpio_phy.h"
#include "nthw_mac_pcs.h"
#include "nthw_sdc.h"

#include "nthw_si5340.h"


#endif	/* __NTHW_CORE_H__ */
