/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(C) 2020 Marvell International Ltd.
 */

#include <rte_trace_point_register.h>

#include "test_trace.h"

RTE_TRACE_POINT_REGISTER(app_dpdk_test_tp, app.dpdk.test.tp)
RTE_TRACE_POINT_REGISTER(app_dpdk_test_fp, app.dpdk.test.fp)
