/* SPDX-License-Identifier: BSD-3-Clause */

#ifndef _RTE_OS_SHIM_
#define _RTE_OS_SHIM_

#include <rte_os.h>

/**
 * @file
 * @internal
 * Provides semi-standard OS facilities by convenient names.
 */

#endif /* _RTE_OS_SHIM_ */
