/* SPDX-License-Identifier: BSD-3-Clause */
/* Copyright (c) Amazon.com, Inc. or its affiliates.
 * All rights reserved.
 */

#ifndef ENA_PLAT_H_
#define ENA_PLAT_H_

#include <ena_plat_dpdk.h>

#endif /* ENA_PLAT_H_ */
