/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2015-2020 Beijing WangXun Technology Co., Ltd.
 */

#ifndef _TXGBE_H_
#define _TXGBE_H_

#include "txgbe_type.h"
#include "txgbe_mng.h"
#include "txgbe_mbx.h"
#include "txgbe_eeprom.h"
#include "txgbe_phy.h"
#include "txgbe_hw.h"
#include "txgbe_vf.h"
#include "txgbe_dcb.h"

#endif /* _TXGBE_H_ */
