/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(C) 2022 Marvell International Ltd.
 */

#ifndef _CMDLINE_CMAN_H_
#define _CMDLINE_CMAN_H_

extern cmdline_parse_inst_t cmd_show_port_cman_capa;
extern cmdline_parse_inst_t cmd_show_port_cman_config;
extern cmdline_parse_inst_t cmd_set_port_cman_config;

#endif /* _CMDLINE_CMAN_H_ */
