/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 2023 Corigine, Inc.
 * All rights reserved.
 */

#include "nfp_vdpa_log.h"

RTE_LOG_REGISTER_SUFFIX(nfp_logtype_core, core, NOTICE);
RTE_LOG_REGISTER_SUFFIX(nfp_logtype_vdpa, vdpa, NOTICE);
