/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 2024 Robin Jarry
 * All rights reserved.
 */

#ifndef _RTE_IP_H_
#define _RTE_IP_H_

#include <rte_ip4.h>
#include <rte_ip6.h>

#endif /* _RTE_IP_H_ */
