/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(C) 2021 Marvell.
 */

#ifndef _ROC_PRIV_H_
#define _ROC_PRIV_H_

/* Utils */
#include "roc_util_priv.h"

/* Mbox */
#include "roc_mbox_priv.h"

/* NPA */
#include "roc_npa_priv.h"

/* Dev */
#include "roc_dev_priv.h"

/* idev */
#include "roc_idev_priv.h"

/* NIX */
#include "roc_nix_priv.h"

/* NPC */
#include "roc_npc_priv.h"

/* SSO */
#include "roc_sso_priv.h"

/* TIM */
#include "roc_tim_priv.h"

/* BPHY CGX */
#include "roc_bphy_cgx_priv.h"

/* CPT */
#include "roc_cpt_priv.h"

/* NIX Inline dev */
#include "roc_nix_inl_priv.h"

/* DPI */
#include "roc_dpi_priv.h"

/* MCS */
#include "roc_mcs_priv.h"

/* REE */
#include "roc_ree_priv.h"

/* ML */
#include "roc_ml_priv.h"

/* RVU LF */
#include "roc_rvu_lf_priv.h"
#endif /* _ROC_PRIV_H_ */
