/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2001-2020 Intel Corporation
 */

#ifndef _IXGBE_HV_VF_H_
#define _IXGBE_HV_VF_H_

#include "ixgbe_type.h"

s32 ixgbevf_hv_init_ops_vf(struct ixgbe_hw *hw);

#endif /* _IXGBE_HV_VF_H_ */
