/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2024 Napatech A/S
 */

/*
 * nthw_fpga_reg_defs_tx_rpl.h
 *
 * Auto-generated file - do *NOT* edit
 *
 */

#ifndef _NTHW_FPGA_REG_DEFS_TX_RPL_
#define _NTHW_FPGA_REG_DEFS_TX_RPL_

/* TX_RPL */
#define NTHW_MOD_TX_RPL (0x1095dfbbUL)

#endif	/* _NTHW_FPGA_REG_DEFS_TX_RPL_ */

/*
 * Auto-generated file - do *NOT* edit
 */
