/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 2023 Red Hat, Inc.
 */

#ifndef VIRTIO_NET_CTRL_H
#define VIRTIO_NET_CTRL_H

int virtio_net_ctrl_handle(struct virtio_net *dev);

#endif /* VIRTIO_NET_CTRL_H */
