/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2024 Napatech A/S
 */

/*
 * nthw_fpga_reg_defs_flm.h
 *
 * Auto-generated file - do *NOT* edit
 *
 */

#ifndef _NTHW_FPGA_REG_DEFS_FLM_
#define _NTHW_FPGA_REG_DEFS_FLM_

/* FLM */
#define NTHW_MOD_FLM (0xe7ba53a4UL)
#define FLM_BUF_CTRL (0x2d7ba0b7UL)
#define FLM_BUF_CTRL_INF_AVAIL (0x55993c46UL)
#define FLM_BUF_CTRL_LRN_FREE (0x60b97a24UL)
#define FLM_BUF_CTRL_STA_AVAIL (0x44abe7c4UL)
#define FLM_CONTROL (0xdbb393a2UL)
#define FLM_CONTROL_CALIB_RECALIBRATE (0xd787824aUL)
#define FLM_CONTROL_CRCRD (0x35ba7f13UL)
#define FLM_CONTROL_CRCWR (0xbc193e07UL)
#define FLM_CONTROL_EAB (0xa09637c2UL)
#define FLM_CONTROL_ENABLE (0x5d80d95eUL)
#define FLM_CONTROL_INIT (0x69b06e85UL)
#define FLM_CONTROL_LDS (0xb880d8faUL)
#define FLM_CONTROL_LFS (0x8ab6ba78UL)
#define FLM_CONTROL_LIS (0xd2ea6b7UL)
#define FLM_CONTROL_PDS (0xadbc82eeUL)
#define FLM_CONTROL_PIS (0x1812fca3UL)
#define FLM_CONTROL_RBL (0x756afcf3UL)
#define FLM_CONTROL_RDS (0xae385680UL)
#define FLM_CONTROL_RIS (0x1b9628cdUL)
#define FLM_CONTROL_SPLIT_SDRAM_USAGE (0x71e7a8a4UL)
#define FLM_CONTROL_UDS (0xab774005UL)
#define FLM_CONTROL_UIS (0x1ed93e48UL)
#define FLM_CONTROL_WPD (0x58ec6f9UL)
#define FLM_INF_DATA (0xba19f6ccUL)
#define FLM_INF_DATA_BYTES (0x480dab67UL)
#define FLM_INF_DATA_BYT_A (0xb9920f4UL)
#define FLM_INF_DATA_BYT_B (0x9290714eUL)
#define FLM_INF_DATA_CAUSE (0x94e9716UL)
#define FLM_INF_DATA_EOR (0xd81a867eUL)
#define FLM_INF_DATA_ID (0x23a04258UL)
#define FLM_INF_DATA_PACKETS (0x33a4ab9eUL)
#define FLM_INF_DATA_PCK_A (0x3967b7a0UL)
#define FLM_INF_DATA_PCK_B (0xa06ee61aUL)
#define FLM_INF_DATA_RTX_A (0x900996cfUL)
#define FLM_INF_DATA_RTX_B (0x900c775UL)
#define FLM_INF_DATA_TCP_A (0xdc945df1UL)
#define FLM_INF_DATA_TCP_B (0x459d0c4bUL)
#define FLM_INF_DATA_TS (0x5f1fab83UL)
#define FLM_LOAD_APS (0x4e7601e5UL)
#define FLM_LOAD_APS_APS (0x504ad426UL)
#define FLM_LOAD_BIN (0xb4367a7dUL)
#define FLM_LOAD_BIN_BIN (0x274bb543UL)
#define FLM_LOAD_LPS (0x46ae92b6UL)
#define FLM_LOAD_LPS_LPS (0x394526b5UL)
#define FLM_LRN_CTRL (0x11050e66UL)
#define FLM_LRN_CTRL_FREE (0x4193813dUL)
#define FLM_LRN_DATA (0xbed48c7fUL)
#define FLM_LRN_DATA_ADJ (0x130ed7UL)
#define FLM_LRN_DATA_COLOR (0x33d0a7f2UL)
#define FLM_LRN_DATA_DSCP (0x5eab148eUL)
#define FLM_LRN_DATA_ENT (0x7fa73e2UL)
#define FLM_LRN_DATA_EOR (0xf782e796UL)
#define FLM_LRN_DATA_FILL (0x768aba23UL)
#define FLM_LRN_DATA_FT (0x4e9221cfUL)
#define FLM_LRN_DATA_FT_MBR (0x48f095acUL)
#define FLM_LRN_DATA_FT_MISS (0xea062e35UL)
#define FLM_LRN_DATA_GFI (0xafa1415dUL)
#define FLM_LRN_DATA_ID (0xd4bd2d64UL)
#define FLM_LRN_DATA_KID (0x5f92d84bUL)
#define FLM_LRN_DATA_MBR_ID1 (0x9931440eUL)
#define FLM_LRN_DATA_MBR_ID2 (0x3815b4UL)
#define FLM_LRN_DATA_MBR_ID3 (0x773f2522UL)
#define FLM_LRN_DATA_MBR_ID4 (0xe95bb081UL)
#define FLM_LRN_DATA_NAT_EN (0x9f4035a4UL)
#define FLM_LRN_DATA_NAT_IP (0xc9fa47cbUL)
#define FLM_LRN_DATA_NAT_PORT (0x5f8f57d0UL)
#define FLM_LRN_DATA_NOFI (0x3d36f27bUL)
#define FLM_LRN_DATA_OP (0x983d5e9fUL)
#define FLM_LRN_DATA_PRIO (0xf7f55b0eUL)
#define FLM_LRN_DATA_PROT (0x2bca3564UL)
#define FLM_LRN_DATA_QFI (0xb70a9e9fUL)
#define FLM_LRN_DATA_QW0 (0xcd0a7417UL)
#define FLM_LRN_DATA_QW4 (0xca67b00eUL)
#define FLM_LRN_DATA_RATE (0x5c250baeUL)
#define FLM_LRN_DATA_RQI (0xb0cfa450UL)
#define FLM_LRN_DATA_SCRUB_PROF (0xc3730f6bUL)
#define FLM_LRN_DATA_SIZE (0x740910fdUL)
#define FLM_LRN_DATA_STAT_PROF (0xded894eaUL)
#define FLM_LRN_DATA_SW8 (0xc055284bUL)
#define FLM_LRN_DATA_SW9 (0xb75218ddUL)
#define FLM_LRN_DATA_TAU (0xea8196fcUL)
#define FLM_LRN_DATA_TEID (0xf62ca024UL)
#define FLM_LRN_DATA_TTL (0xb95fd828UL)
#define FLM_LRN_DATA_VOL_IDX (0xabc86ffbUL)
#define FLM_PRIO (0x5ed7bcbeUL)
#define FLM_PRIO_FT0 (0x9ef34f69UL)
#define FLM_PRIO_FT1 (0xe9f47fffUL)
#define FLM_PRIO_FT2 (0x70fd2e45UL)
#define FLM_PRIO_FT3 (0x7fa1ed3UL)
#define FLM_PRIO_LIMIT0 (0xcce9cfe8UL)
#define FLM_PRIO_LIMIT1 (0xbbeeff7eUL)
#define FLM_PRIO_LIMIT2 (0x22e7aec4UL)
#define FLM_PRIO_LIMIT3 (0x55e09e52UL)
#define FLM_PST_CTRL (0x3e2b004bUL)
#define FLM_PST_CTRL_ADR (0xfa8565c1UL)
#define FLM_PST_CTRL_CNT (0xea8dfc10UL)
#define FLM_PST_DATA (0x91fa8252UL)
#define FLM_PST_DATA_BP (0x5f53d50fUL)
#define FLM_PST_DATA_PP (0x27a7a5dcUL)
#define FLM_PST_DATA_TP (0x43cb60d8UL)
#define FLM_RCP_CTRL (0x8041d9eeUL)
#define FLM_RCP_CTRL_ADR (0xb1a39fefUL)
#define FLM_RCP_CTRL_CNT (0xa1ab063eUL)
#define FLM_RCP_DATA (0x2f905bf7UL)
#define FLM_RCP_DATA_A (0xec41ba43UL)
#define FLM_RCP_DATA_AUTO_IPV4_MASK (0xa7818261UL)
#define FLM_RCP_DATA_B (0x7548ebf9UL)
#define FLM_RCP_DATA_BYT_DYN (0x28334583UL)
#define FLM_RCP_DATA_BYT_OFS (0x8a3ac825UL)
#define FLM_RCP_DATA_IPN (0x94f5d891UL)
#define FLM_RCP_DATA_ITF (0xfe4295a7UL)
#define FLM_RCP_DATA_KID (0xeca44cf9UL)
#define FLM_RCP_DATA_LOOKUP (0x5d0f2e84UL)
#define FLM_RCP_DATA_MASK (0xd97a1393UL)
#define FLM_RCP_DATA_OPN (0x9078a423UL)
#define FLM_RCP_DATA_QW0_DYN (0x28bd732dUL)
#define FLM_RCP_DATA_QW0_OFS (0x8ab4fe8bUL)
#define FLM_RCP_DATA_QW0_SEL (0x39adfaa9UL)
#define FLM_RCP_DATA_QW4_DYN (0xdd3dd5edUL)
#define FLM_RCP_DATA_QW4_OFS (0x7f34584bUL)
#define FLM_RCP_DATA_SW8_DYN (0x8f5229c5UL)
#define FLM_RCP_DATA_SW8_OFS (0x2d5ba463UL)
#define FLM_RCP_DATA_SW8_SEL (0x9e42a041UL)
#define FLM_RCP_DATA_SW9_DYN (0xb2320075UL)
#define FLM_RCP_DATA_SW9_OFS (0x103b8dd3UL)
#define FLM_RCP_DATA_TXPLM (0xbe56af35UL)
#define FLM_SCAN (0xee586089UL)
#define FLM_SCAN_I (0xe22d4ee5UL)
#define FLM_SCRUB (0x690c7a66UL)
#define FLM_SCRUB_I (0xc6a9dfe8UL)
#define FLM_SCRUB_CTRL (0xc647074aUL)
#define FLM_SCRUB_CTRL_ADR (0x3d584aa4UL)
#define FLM_SCRUB_CTRL_CNT (0x2d50d375UL)
#define FLM_SCRUB_DATA (0x69968553UL)
#define FLM_SCRUB_DATA_DEL (0xc96d19b1UL)
#define FLM_SCRUB_DATA_INF (0xc294ba37UL)
#define FLM_SCRUB_DATA_R (0xc3ff3ba5UL)
#define FLM_SCRUB_DATA_T (0x2a9c9e90UL)
#define FLM_STAT (0xa532c06UL)
#define FLM_STAT_I (0x215c23d1UL)
#define FLM_STATUS (0x10f57a96UL)
#define FLM_STATUS_CACHE_BUFFER_CRITICAL (0x39f44c00UL)
#define FLM_STATUS_CALIB_FAIL (0xb59e2af6UL)
#define FLM_STATUS_CALIB_SUCCESS (0x9c031e4fUL)
#define FLM_STATUS_CRCERR (0xcd852b33UL)
#define FLM_STATUS_CRITICAL (0xe9e1b478UL)
#define FLM_STATUS_EFT_BP (0xf60fc391UL)
#define FLM_STATUS_EFT_EVICT_BP (0xf4dd216fUL)
#define FLM_STATUS_IDLE (0xd02fd0e7UL)
#define FLM_STATUS_INITDONE (0xdd6bdbd8UL)
#define FLM_STATUS_PANIC (0xf676390fUL)
#define FLM_STAT_AUL_DONE (0x747800bbUL)
#define FLM_STAT_AUL_DONE_CNT (0x6b744caeUL)
#define FLM_STAT_AUL_FAIL (0xe272cb59UL)
#define FLM_STAT_AUL_FAIL_CNT (0x697b6247UL)
#define FLM_STAT_AUL_IGNORE (0x49aa46f4UL)
#define FLM_STAT_AUL_IGNORE_CNT (0x37721e3bUL)
#define FLM_STAT_CSH_HIT (0xd0dca28cUL)
#define FLM_STAT_CSH_HIT_CNT (0xc7a41ba4UL)
#define FLM_STAT_CSH_MISS (0xbc219a9fUL)
#define FLM_STAT_CSH_MISS_CNT (0x4a66e57eUL)
#define FLM_STAT_CSH_UNH (0x9f625827UL)
#define FLM_STAT_CSH_UNH_CNT (0x79b8ac91UL)
#define FLM_STAT_CUC_MOVE (0x381862c0UL)
#define FLM_STAT_CUC_MOVE_CNT (0x868e1261UL)
#define FLM_STAT_CUC_START (0x617a68acUL)
#define FLM_STAT_CUC_START_CNT (0xbf953f79UL)
#define FLM_STAT_FLOWS (0x3072544fUL)
#define FLM_STAT_FLOWS_CNT (0x829b2631UL)
#define FLM_STAT_INF_DONE (0x63dff05cUL)
#define FLM_STAT_INF_DONE_CNT (0xf78de916UL)
#define FLM_STAT_INF_SKIP (0x8b84458aUL)
#define FLM_STAT_INF_SKIP_CNT (0xc0b9dd7dUL)
#define FLM_STAT_LRN_DONE (0x67128aefUL)
#define FLM_STAT_LRN_DONE_CNT (0xd81588feUL)
#define FLM_STAT_LRN_FAIL (0xf118410dUL)
#define FLM_STAT_LRN_FAIL_CNT (0xda1aa617UL)
#define FLM_STAT_LRN_IGNORE (0x9a10a7f0UL)
#define FLM_STAT_LRN_IGNORE_CNT (0x11c0f6a7UL)
#define FLM_STAT_PCK_DIS (0x55e23053UL)
#define FLM_STAT_PCK_DIS_CNT (0x9b13b227UL)
#define FLM_STAT_PCK_HIT (0xc29c5c94UL)
#define FLM_STAT_PCK_HIT_CNT (0xee930443UL)
#define FLM_STAT_PCK_MISS (0xaf5f4237UL)
#define FLM_STAT_PCK_MISS_CNT (0x7421a5baUL)
#define FLM_STAT_PCK_UNH (0x8d22a63fUL)
#define FLM_STAT_PCK_UNH_CNT (0x508fb376UL)
#define FLM_STAT_PRB_DONE (0x3bc46ef0UL)
#define FLM_STAT_PRB_DONE_CNT (0xcb3bc80dUL)
#define FLM_STAT_PRB_IGNORE (0xf02dd3d9UL)
#define FLM_STAT_PRB_IGNORE_CNT (0x588d1667UL)
#define FLM_STAT_REL_DONE (0xe192aabdUL)
#define FLM_STAT_REL_DONE_CNT (0xbe04b769UL)
#define FLM_STAT_REL_IGNORE (0x29f33e6eUL)
#define FLM_STAT_REL_IGNORE_CNT (0x99a6a156UL)
#define FLM_STAT_STA_DONE (0x500f2e87UL)
#define FLM_STAT_STA_DONE_CNT (0xbf276bbdUL)
#define FLM_STAT_TUL_DONE (0x40233ff4UL)
#define FLM_STAT_TUL_DONE_CNT (0xffcfd642UL)
#define FLM_STAT_UNL_DONE (0xe950f75eUL)
#define FLM_STAT_UNL_DONE_CNT (0xe9bdd9a2UL)
#define FLM_STAT_UNL_IGNORE (0x497abbcaUL)
#define FLM_STAT_UNL_IGNORE_CNT (0x1ba2b435UL)
#define FLM_STA_DATA (0x89c92817UL)
#define FLM_STA_DATA_EOR (0x90b004d5UL)
#define FLM_STA_DATA_ID (0x3b14a2ffUL)
#define FLM_STA_DATA_LDS (0xb92d607UL)
#define FLM_STA_DATA_LFS (0x39a4b485UL)
#define FLM_STA_DATA_LIS (0xbe3ca84aUL)
#define FLM_STA_DATA_PDS (0x1eae8c13UL)
#define FLM_STA_DATA_PIS (0xab00f25eUL)
#define FLM_STA_DATA_RDS (0x1d2a587dUL)
#define FLM_STA_DATA_RIS (0xa8842630UL)
#define FLM_STA_DATA_UDS (0x18654ef8UL)
#define FLM_STA_DATA_UIS (0xadcb30b5UL)
#define FLM_TRSWIN (0xff97af47UL)
#define FLM_TRSWIN_S (0x41ed83b1UL)
#define FLM_TRTWIN (0x624097feUL)
#define FLM_TRTWIN_T (0xc28c26aaUL)

#endif	/* _NTHW_FPGA_REG_DEFS_FLM_ */

/*
 * Auto-generated file - do *NOT* edit
 */
