/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2022 Loongson Technology Corporation Limited
 */

#ifndef RTE_POWER_INTRINSIC_LOONGARCH_H
#define RTE_POWER_INTRINSIC_LOONGARCH_H

#include <rte_common.h>

#include "generic/rte_power_intrinsics.h"

#endif /* RTE_POWER_INTRINSIC_LOONGARCH_H */
