/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(C) 2021 Marvell.
 */

#ifndef _CN9K_CRYPTODEV_H_
#define _CN9K_CRYPTODEV_H_

/* Marvell OCTEON CN9K Crypto PMD device name */
#define CRYPTODEV_NAME_CN9K_PMD crypto_cn9k

extern uint8_t cn9k_cryptodev_driver_id;

#endif /* _CN9K_CRYPTODEV_H_ */
