/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2020 Broadcom
 * All rights reserved.
 */

#ifndef _BCMFS_SYM_CAPABILITIES_H_
#define _BCMFS_SYM_CAPABILITIES_H_

/*
 * Get capabilities list for the device
 *
 */
const struct rte_cryptodev_capabilities *bcmfs_sym_get_capabilities(void);

#endif /* _BCMFS_SYM_CAPABILITIES_H__ */
