/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2018 Chelsio Communications.
 * All rights reserved.
 */

#ifndef __T4VF_HW_H
#define __T4VF_HW_H

#define T4VF_PL_BASE_ADDR	0x0200
#define T4VF_CIM_BASE_ADDR	0x0300
#define T4VF_MBDATA_BASE_ADDR	0x0240
#define T6VF_MBDATA_BASE_ADDR	0x0280

#define NUM_CIM_VF_MAILBOX_DATA_INSTANCES NUM_CIM_PF_MAILBOX_DATA_INSTANCES
#endif /* __T4VF_HW_H */
