/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2016 Cavium, Inc
 * Copyright(c) 2022 StarFive
 * Copyright(c) 2022 SiFive
 * Copyright(c) 2022 Semihalf
 */

#ifndef RTE_IO_RISCV_H
#define RTE_IO_RISCV_H

#include "generic/rte_io.h"

#endif /* RTE_IO_RISCV_H */
