/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(C) 2021 Marvell.
 */
#ifndef _CNXK_UTILS_H_
#define _CNXK_UTILS_H_

#include "roc_platform.h"

int __roc_api roc_nix_tm_err_to_rte_err(int errorcode);

#endif /* _CNXK_UTILS_H_ */
