/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2018 Intel Corporation
 */

#ifndef _BPF_CMD_H_
#define _BPF_CMD_H_

#ifdef RTE_LIB_BPF

 /* BPF CLI */
extern cmdline_parse_inst_t cmd_operate_bpf_ld_parse;
extern cmdline_parse_inst_t cmd_operate_bpf_unld_parse;

#endif /* RTE_LIB_BPF */

#endif /* _BPF_CMD_H_ */
