/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2024 Napatech A/S
 */

/*
 * nthw_fpga_reg_defs_rpp_lr.h
 *
 * Auto-generated file - do *NOT* edit
 *
 */

#ifndef _NTHW_FPGA_REG_DEFS_RPP_LR_
#define _NTHW_FPGA_REG_DEFS_RPP_LR_

/* RPP_LR */
#define NTHW_MOD_RPP_LR (0xba7f945cUL)
#define RPP_LR_IFR_RCP_CTRL (0xce88594UL)
#define RPP_LR_IFR_RCP_CTRL_ADR (0x4b4cc068UL)
#define RPP_LR_IFR_RCP_CTRL_CNT (0x5b4459b9UL)
#define RPP_LR_IFR_RCP_DATA (0xa339078dUL)
#define RPP_LR_IFR_RCP_DATA_IPV4_DF_DROP (0xee1d681fUL)
#define RPP_LR_IFR_RCP_DATA_IPV4_EN (0xfe386131UL)
#define RPP_LR_IFR_RCP_DATA_IPV6_DROP (0x41f324ffUL)
#define RPP_LR_IFR_RCP_DATA_IPV6_EN (0x5431a9baUL)
#define RPP_LR_IFR_RCP_DATA_MTU (0x871a2322UL)
#define RPP_LR_RCP_CTRL (0xf3395d47UL)
#define RPP_LR_RCP_CTRL_ADR (0x4916a944UL)
#define RPP_LR_RCP_CTRL_CNT (0x591e3095UL)
#define RPP_LR_RCP_DATA (0x5ce8df5eUL)
#define RPP_LR_RCP_DATA_EXP (0x578ca035UL)

#endif	/* _NTHW_FPGA_REG_DEFS_RPP_LR_ */

/*
 * Auto-generated file - do *NOT* edit
 */
