/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2024 Napatech A/S
 */

/*
 * nthw_fpga_reg_defs_ifr.h
 *
 * Auto-generated file - do *NOT* edit
 *
 */

#ifndef _NTHW_FPGA_REG_DEFS_IFR_
#define _NTHW_FPGA_REG_DEFS_IFR_

/* IFR */
#define NTHW_MOD_IFR (0x9b01f1e6UL)
#define IFR_COUNTERS_CTRL (0x92ba13e6UL)
#define IFR_COUNTERS_CTRL_ADR (0xecdeeda8UL)
#define IFR_COUNTERS_CTRL_CNT (0xfcd67479UL)
#define IFR_COUNTERS_DATA (0x3d6b91ffUL)
#define IFR_COUNTERS_DATA_DROP (0x3ee57cc0UL)
#define IFR_DF_BUF_CTRL (0xf60805e4UL)
#define IFR_DF_BUF_CTRL_AVAILABLE (0x158f09c3UL)
#define IFR_DF_BUF_CTRL_MTU_PROFILE (0x7cb4bc5aUL)
#define IFR_DF_BUF_DATA (0x59d987fdUL)
#define IFR_DF_BUF_DATA_FIFO_DAT (0xdbfdf650UL)
#define IFR_RCP_CTRL (0xc6dfc47eUL)
#define IFR_RCP_CTRL_ADR (0x68600d59UL)
#define IFR_RCP_CTRL_CNT (0x78689488UL)
#define IFR_RCP_DATA (0x690e4667UL)
#define IFR_RCP_DATA_IPV4_DF_DROP (0xbfd1ca18UL)
#define IFR_RCP_DATA_IPV4_EN (0xb48ee13aUL)
#define IFR_RCP_DATA_IPV6_DROP (0x4fbf34a2UL)
#define IFR_RCP_DATA_IPV6_EN (0x1e8729b1UL)
#define IFR_RCP_DATA_MTU (0xa436ee13UL)

#endif	/* _NTHW_FPGA_REG_DEFS_IFR_ */

/*
 * Auto-generated file - do *NOT* edit
 */
